#include "DemoApp.h"
#include "TestLib.h"

/* ------------ Demo-App v1.1 ----------------- */
/* Autor: Maxim Kpper							*/
/* -------------------------------------------- */

	volatile int iProgStatus = 0;		// 0 = Programm noch nicht gestartet
										// 1 = Programm gestartet, luft
										// ---------------- WICHTIG: -------------------------
										// volatile, sonst ndert sich progstatus nicht in der ISR!
									
	char cStatus = 10;					// cStatus beinhaltet den aktuellen Status des Programmes
										// wird zur Synchronisation bentigt
	int iMaxStat = 255;					// Maximaler Status, da 8 mgliche zustnde der LEDS
									
									
	int iDelayMs = 10*DELAY;				// Wartezeit zur Synchronisation in ms (nicht 100% genau,
										// da Taktfrequenz statt durch 1000 (auf kHz) durch 1024 geteilt wurde)
										// Maximum	:	~30ms
										
	int iMaxTimeMs = 20*DELAY;			// Maximale Zeit eines Schleifendurchlaufs des Hauptprogramms
										// Angabe in ms
										// Maximum	:	~30ms
										
	int iDebugFlag = 0;					// 1	: 	Programm startet automatisch
										// 0	:	Programm wartet auf Startsignal
										// Startsignal	:	Startknopf od. externes Startsignal
										// Fr Simulator 1 einsetzen, dann wird auerdem das Delay dekativiert!
										// ---------------- WICHTIG: -------------------------
										// IM DEBUG-MODUS WERDEN DIE SYNCHRONISATION 
										// UND DIE SELBSTTESTS DEAKTIVIERT!
										
	int iTestFlag = 0;					// 1	:	Selbsttests werden durchgefhrt
										// 0	:	keine Selbsttests
										
	int iSyncFlag = 1;						// Flag ob Synchronisation durchgefhrt werden soll
										// 0		: 	deaktiviert
										// 1		: 	aktiviert
										
	int iOn = 0xFF;						// alle Lampen an	/ Port als Ausgang deklarieren
	int iOff = 0x00;					// alle Lampen aus	/ Port als Eingang deklarieren
	
	int iWDStatusFlag;					// Status des Watchdog-Tests
	
	char cReceived;


/* ---------------------------------------------------------------------------------- */
/* ------------------------------- Hauptmethode ------------------------------------- */
/* ---------------------------------------------------------------------------------- */
int main(void)
{	
	// System-Clock auf 2 MHz einstellen
	//CLKPR |= (1<<CLKPCE);
	//CLKPR = (1<<CLKPS1);
		
	iWDStatusFlag = eepromRead(iSpeicher3);
	
	// Watchdog-Routine, wenn Watchdog einen Systemreset durchgefhrt hat
	if ((MCUSR & (1 << WDRF))){
		MCUSR = 0;
		wdt_disable();
		if(iWDStatusFlag != 1){
			DDRB = iOn;
			DDRD = iOn;
			wdError();
		}
	}
	
	// Ruft den Watchdog-Test auf wenn er noch nicht aufgerufen wurde und Tests aktiviert sind
	if(iWDStatusFlag != 1 && iTestFlag){
		startTest(54);
	}
	else if(iTestFlag){
		eepromWrite(iSpeicher3, 0);
		iTestStat = 0;
	}

	// Ruft die Methode zum Programmstart auf
	programmInit();
	
	// Prfen ob Debugmodus aktiviert, wenn nein Synchronisation starten und Tests durchfhren
	if(!iDebugFlag){
		// Falls Tests gewnscht werden
		if(iTestFlag){
			// Beim Start einmalig alle Selbsttests durchfhren
			initTest();
		}
		// Status-LEDs setzen, jenachdem ob Synchronisation erwnscht oder nicht
		if(iSyncFlag){
			PORTD = iOn;	
		}
		else if(!iSyncFlag){
			PORTD = 0x33;
		}
	}
	
	// Schleife vor dem eigentlichen Programmstart. 
	// Fhrt Selbsttests der Reihe nach aus. 
	// Wartet auf externes Signal (Button-Aktion oder Start-Signal) 
	while(iProgStatus != 1){
		doTest();								// Routine zur Testauswahl aufrufen
		wdt_reset();							// Watchdog zurck setzen
		if((UCSRA & (1<<RXC)) && iSyncFlag){		// Prft ob externes Signal vorliegt
			if(UDR == 'x'){					// Im Falle das externes Signal ein x ist..
				programmStart();				// .. fhre Programmstart aus
			}
		}
	}
	
	PORTD = iOff;
	DDRB = iOn; 								// PortB als Ausgang schalten
	PORTB = iOff;								// Initialwert dem PortB zuweisen
	
	// Timer fr die Echtzeit-berwachung
	OCR0A = iMaxTimeMs;							// Ldt das Vergleichsregister
	TCCR0A = (1<<CS00)|(1<<CS02)|(1<<WGM01);	// Initialisiert den Timer und startet ihn
	
	// Hauptprogamm-Endlosschleife
	// wird nur durch einen Fehlerfall verlassen
	while(1){	

		TCNT0 = 0;								// Setzt den Timer auf 0 zurck
												// .. Wenn ja, Status zurck auf 0 setzen	    

		if(cStatus < iMaxStat)					// Prft ob der Status den hchsten Wert erreicht hat..
			cStatus++;							// .. wenn nein, Status inkrementieren
		else  							
			cStatus = 0;
		
		PORTB = cStatus;						// Status ber die LEDs ausgeben.
				
		if(iSyncFlag){								// Falls Syncronisationsflag gesetzt..	
			doSync();							// ..Sychronisation ausfhren
		}
		
		// fr den Debug-Modus die Tests deaktivieren
		if(iTestFlag){
			doTest();							// Selbsttest-Auswahl aufrufen
		}
		
		wdt_reset();							// Watchdog zurck setzen
	}
	
	return 0;
	
}


/* ---------------------------------------------------------------------------------- */
/* ----------------------- Aufruf der Initialisierung ------------------------------- */
/* ---------------------------------------------------------------------------------- */
void programmInit(void){
	// Wennn Autostart aktiviert ist, setze Progstatus auf 1 zum Starten des Programmes
	if(iDebugFlag){
		iProgStatus = 1;
		iSyncFlag = 0;
		iTestFlag = 0;
	}
	
	eepromOverwrite();

	// Ruft die Initialisierungsmethode auf
	initializer();
	
	// Initialisiert den Watchdog und stellt ihn auf den gewnschten Wert
	// WDTO_60MS entspricht 60mS
	// WDTO_30MS entspricht 30mS
	// WDTO_1S entspricht 1S
	wdt_enable (WDTO_60MS);
}

/* ---------------------------------------------------------------------------------- */
/* ------------------------- Routine zum Programmstart ------------------------------ */
/* ---------------------------------------------------------------------------------- */
void programmStart(void){
	iProgStatus = 1;							// Programmstatus-Variable auf 1 stellen
	PCMSK1 = ~PINB_MASK;						// Pin-Change-Interrupt deaktivieren
	EIMSK = (0<<7);
}


/* ---------------------------------------------------------------------------------- */
/* -------------------------- globale Initialisierung ------------------------------- */
/* ---------------------------------------------------------------------------------- */
void initializer()
{
	//Init USART
	USARTinit();
	
	TIMSK0 = (1<<OCIE0A);						// Timer-Interrupts aktivieren
	TIMSK2 = (1<<OCIE2A);
	
	if(!iDebugFlag){
		// PortB-Pin 4 als Eingang schalten & Pull-Up Widerstnde schalten
		DDRB = iOff;
		PORTB = 0x10;
		
		// PortD-Pin als Ausgang fr die Fehlermeldung schalten
		DDRD = iOn;
		PORTD = iOff;
	
		// Pin-Change-Interrupt fr PortB schalten
		PCMSK1 = PINB_MASK;
		EIMSK = (1<<7);
	}
}

/* ---------------------------------------------------------------------------------- */
/* -------------------------- Taster-Interrupt-Routine ------------------------------ */
/* ---------------------------------------------------------------------------------- */
SIGNAL(PCINT1_vect)
{
    PinChangeInterrupt();  
}


/* ---------------------------------------------------------------------------------- */
/* -------------------- Ausfhrende Routine der Taster-ISR -------------------------- */
/* ---------------------------------------------------------------------------------- */
void PinChangeInterrupt(void)
{
    char cbuttons;
		
    cbuttons = (~PINB) & PINB_MASK;

	// Output virtual keys
	if (cbuttons & (1<<BUTTON_O)){				// berprfe ob gedrckte Taste = Taster war
		programmStart();						// wenn ja, programmstart ausfhren
		if(iSyncFlag){							// und falls gewnscht dem anderen System das Startzeichen senden
			sendChar('x');
		}
	}
	
	EIFR = (0<<PCIF1);    						// Lsche Pinchange-Interrupt-Flag
}


/* ---------------------------------------------------------------------------------- */
/* --------------------------------- Daten senden ----------------------------------- */
/* ---------------------------------------------------------------------------------- */
void sendChar(char data)
{
	// Darauf warten, dass das Senderegister leer wird und bereit fr neue Daten zum Senden ist.
	while(!(UCSRA & (1<<UDRE)));

    // Daten ins Senderegister schieben
    UDR = data;

	// Warte darauf, dass Daten gesendet werden
	while(!(UCSRA & (1<<TXC)));
}


/* ---------------------------------------------------------------------------------- */
/* -------------------- Serielle Schnittstelle initialisieren ----------------------- */
/* ---------------------------------------------------------------------------------- */
void USARTinit()
{
	
    // Serielle Schnittstelle als Sender und Empfnger initialisieren
    UCSRB = (1<<RXEN)|(1<<TXEN)|(0<<RXCIE)|(0<<UDRIE);

    // Asynchroner Modus mit 8 Bits, keine Parity und einem Stop-Bit einstellen
    UCSRC = (0<<UMSEL)|(0<<UPM0)|(0<<USBS)|(3<<UCSZ0)|(0<<UCPOL);
 
	// Baudratenwert (Berechnung siehe Header) dem Register zuweisen
	UBRRH = (unsigned long)UBRR_VAL>>8;
	UBRRL = (unsigned long)UBRR_VAL;
	

    // Enable interrupts
    sei(); 
}


/* ---------------------------------------------------------------------------------- */
/* --------------------- Fehlerroutine fr unbekannte Fehler ------------------------ */
/* ---------------------------------------------------------------------------------- */
void stdError(){
	cli();										// Interrupt-Behandlungen deaktivieren	
	wdt_disable();								// Watchdog deaktivieren
	
	PORTB = iOff;
	
	eepromWrite(iSpeicher1, 6);					// Fehlerwert ins Eeprom schreiben
	
	while(1){
		// Fehlercode an PortD ausgeben. Fehlercode laut Tabelle
		PORTD = 0xAA;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		PORTD = iOff;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
	}
}


/* ---------------------------------------------------------------------------------- */
/* ------------------- Fehlerroutine fr Synchronisationsfehler --------------------- */
/* ---------------------------------------------------------------------------------- */
void syncError(){	
	cli();										// Interrupt-Behandlungen deaktivieren
	wdt_disable();								// Watchdog deaktivieren
	
	eepromWrite(iSpeicher1, 4);							// Fehlerwert ins Eeprom schreiben
	eepromWrite(iSpeicher2, cStatus);
		
	PORTB = cStatus;
	
	while(1){
		// Fehlercode an PortD ausgeben. Fehlercode laut Tabelle
		PORTD = 0x66;						
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		PORTD = iOff;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
	}
}


/* ---------------------------------------------------------------------------------- */
/* ----------------------- Fehlerroutine fr Watchdog-Reset ------------------------- */
/* ---------------------------------------------------------------------------------- */
void wdError(){
	cli();										// Interrupt-Behandlungen deaktivieren
	wdt_disable();								// Watchdog deaktivieren
	
	eepromWrite(iSpeicher1, 3);					// Fehlerwert ins Eeprom schreiben
	
	while(1){
		// Fehlercode an PortD ausgeben. Fehlercode laut Tabelle
		PORTD = 0x99;						
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		PORTD = iOff;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
	}
}


/* ---------------------------------------------------------------------------------- */
/* ------- Fehlerroutine fr den Fall eines Ausfalls des anderen Systems ------------ */
/* ---------------------------------------------------------------------------------- */
void ausfError(){	
	cli();										// Interrupt-Behandlungen deaktivieren
	wdt_disable();								// Watchdog deaktivieren
	
	eepromWrite(iSpeicher1, 1);							// Fehlerwert ins Eeprom schreiben
	eepromWrite(iSpeicher2, cStatus);		
	
	PORTB = cStatus;
		
	while(1){
		// Fehlercode an PortD ausgeben. Fehlercode laut Tabelle
		PORTD = iOn;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		PORTD = iOff;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
	}
}

/* ---------------------------------------------------------------------------------- */
/* --------- Fehlerroutine fr den Ablauf des Timers der Hauptschleife -------------- */
/* ---------------------------------------------------------------------------------- */
void zeitError(){
	cli();										// Interrupt-Behandlungen deaktivieren
	wdt_disable();								// Watchdog deaktivieren
	
	eepromWrite(iSpeicher1,2);							// Fehlerwert ins Eeprom schreiben
	
	while(1){
		// Fehlercode an PortD ausgeben. Fehlercode laut Tabelle
		PORTD = iOn;
		PORTB = iOn;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		PORTD = iOff;
		PORTB = iOff;
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
		_delay_loop_2(65535);
	}
}


/* ---------------------------------------------------------------------------------- */
/* ----------------------------- Selbsttest-Aufruf ---------------------------------- */
/* ---------------------------------------------------------------------------------- */
void doTest(){
	selectTest();								// Methode zur Testauswahl aufrufen
	iTestStat++;								// aktuellen Zhler erhhen
	if(iTestStat >= (iTestAnzahl))				// Zhler berschreitet maximale Anzahl an implementierten
		iTestStat = 0;							// Tests = Zhler zurck setzen
}

/* ---------------------------------------------------------------------------------- */
/* -------------------------- Synchronisationsroutine  ------------------------------ */
/* ------------------ empfngt und sendet aktuellen Status -------------------------- */
/* ---------------------------------------------------------------------------------- */
void doSync(){

	// aktuellen Status senden
	if(cStatus <= iMaxStat){
		sendChar(cStatus);
	}
	else
		stdError();
	
	// Watchdog  zurck setzen um keinen Alarm beim Warten auf Statusantwort zu schlagen
	wdt_reset();
	
	_delay_loop_2(4000);
	
	TCNT2 = 0;
	OCR2A = iDelayMs;							// Ldt das Vergleichsregister
	TCCR2A = (1<<CS20)|(1<<CS22)|(1<<WGM21);	// Initialisiert den Timer und startet ihn
	
	// Schleife zur berprfung ob Synchronisation stattgefunden hat
	while (!(UCSRA & (1<<RXC)));				// Prfe ob Daten empfangen wurden		
	
	
	
	cReceived = UDR;
	
	if(cReceived != cStatus){							// Wenn Daten empfangen mit aktuellem Status vergleichen
		syncError();							// Bei ungleichheit Sychronittsverlust melden
	}	
		
	TCCR2A &= ~((1<<CS20)|(1<<CS22));			// Timer anhalten
}


/* ---------------------------------------------------------------------------------- */
/* -------------------------- Timer0-Interrupt-Routine ------------------------------ */
/* ---------------------------------------------------------------------------------- */
SIGNAL(TIMER0_COMP_vect)
{
	TCCR0A &= ~((1<<CS00)|(1<<CS02));
	zeitError();
}

/* ---------------------------------------------------------------------------------- */
/* -------------------------- Timer2-Interrupt-Routine ------------------------------ */
/* ---------------------------------------------------------------------------------- */
SIGNAL(TIMER2_COMP_vect)
{
	TCCR2A &= ~((1<<CS20)|(1<<CS22));
	ausfError();
}

/* ---------------------------------------------------------------------------------- */
/* ------------------- Funktion um ins EEEPROM zu schreiben ------------------------- */
/* ---------------------------------------------------------------------------------- */

void eepromWrite(int iSpeicher, int iError){
		eeprom_write_byte((uint8_t*)iSpeicher, iError);
}	

/* ---------------------------------------------------------------------------------- */
/* -------------- berschreibt die Speicherzellen im EEPROM mit 0 ------------------- */
/* -------------- falls diese nicht schon 0 sind. ----------------------------------- */
/* ---------------------------------------------------------------------------------- */
void eepromOverwrite(){
	if(eepromRead(iSpeicher1) != 0){ 
		eepromWrite(iSpeicher1, 0);
	}
	
	if(eepromRead(iSpeicher2) != 0){ 
		eepromWrite(iSpeicher2, 0);
	}
}

/* ---------------------------------------------------------------------------------- */
/* -------------------- Funktion um aus dem EEPROM zu lesen ------------------------- */
/* ---------------------------------------------------------------------------------- */
int eepromRead(int iSpeicher){
	return eeprom_read_byte((uint8_t*)iSpeicher);
}	